package enchanting;

import lejos.robotics.navigation.DifferentialPilot;
import java.util.Hashtable;
import java.io.*;
import lejos.nxt.*;
import lejos.nxt.comm.*;
import lejos.robotics.*;
import lejos.robotics.navigation.*;
import lejos.robotics.localization.*; //PoseProvider;
import lejos.util.Stopwatch;
import javax.microedition.lcdui.Graphics;
import java.util.Random;

public abstract class EnchantingScriptBase extends Thread {
	
	// Here is where the hardcoded pilot and navigator go
	// Here are where sprites are declared
	
	public static Random random = new Random();
	public static Stopwatch stopwatch = new Stopwatch();
	
	static int[] currentInstrument = Sound.PIANO; // default piano
	static double secondsPerBeat = 1.0; // 60 bpm by default
	public static final double MAX_TEMPO = 512.0;

 
 	// Things are initialized here
 

    // HELPER FUNCTIONS
    // ----------------
    
    public static final void delay(float seconds) {
       //Sound.pause( (int)(1000.0f * seconds) );  // time in milliseconds
       try {
           Thread.sleep((int)(1000.0f * seconds));
       } catch (Exception ex){}
    }

    // Added this to rest for a given number of beats
    public static final void restBeats(float beats) {
       delay( (float)(beats * secondsPerBeat) ); // time in seconds
    } 

    // Set the tempo to a given BPM
    public static final void setTempo(double beatsPerMinute) {
    	   if(beatsPerMinute > 0 && beatsPerMinute < MAX_TEMPO) {
	       secondsPerBeat = 60 / beatsPerMinute;
	   } else {
	       secondsPerBeat = 1.0; // 60 bpm
           }
    }

    // Change the tempo by a given BPM
    public static final void changeTempoBy(double bpmOffset) {
           // Calculate the new beats per minute after the offset is added.
           double newBPM = tempo() + bpmOffset;
	   setTempo(newBPM);
    }

    // Return the number of beats per minute
    public static final double tempo() { return 60.0f / secondsPerBeat; }

    // Implemented Mark Hunter's setInstrument function
    public static final void setInstrument(int instrument) {
    	switch(instrument) {
    		case 3: currentInstrument = Sound.FLUTE; break;
    		case 2: currentInstrument = Sound.XYLOPHONE; break;
    		case 1: default: currentInstrument = Sound.PIANO;
    	}
    }
    // Implemented reporter for identifying current instrument on NXT screen.
    // CWB - Changed function to return an int instead of a string
    public static final int instrument() {
    	if(currentInstrument == Sound.PIANO) {
	    	return 1;
		} else if(currentInstrument == Sound.XYLOPHONE) {
			return 2;
		} else if(currentInstrument == Sound.FLUTE) {
			return 3;
		}
	   
		return 0;
    }
    
    // Will play the sample and delay execution until the sample is finished playing
    public static final void playSampleUntilDone(String fileName) {
		int len = playSample(fileName);
		delay(len/100); // delay by (length/100) milliseconds
    }
    
    // Will play the sample until either the program finishes, or the sound finishes, whichever comes first
    public static final int playSample(String fileName) {
		int len = Sound.playSample(new File(fileName), Sound.getVolume());
		return len;
    }

    // Play a note for a certain number of beats
    public static final void playNote(double note, double beats) {
        int middleA = 440;  // hz
        double freq = (middleA / 32) * Math.pow(2.0, (note - 9) / 12);
        double ms = beats * secondsPerBeat * 1000.0f; 
        Sound.playNote(currentInstrument, (int)freq, (int)ms);
    }

    // Set the volume according to the value passed from the FrontEnd
    public static final void setVolume(double vol) {
    	int volume = (int)vol;
		if (volume < 0) {
	   		Sound.setVolume(0);
   	   	} else if (volume > 100) {
	        Sound.setVolume(100);
	   	} else {
 	        Sound.setVolume(volume);
	   	}
    }

    // Change the volume according to the value passed from the FrontEnd
    public static final void changeVolumeBy(double vol) {
	   double newVolume = Sound.getVolume() + vol;
	   setVolume(newVolume);
    }

    // Return the current volume
    public static final int volume() { return Sound.getVolume(); }
    
    public static final int randomRange(int a, int b) {
    	// returns a random number between a and b inclusive
		int low = a;
		int diff = b - a;
		if (diff < 0) {
			low = b;
			diff = -diff;
		}
		return random.nextInt(diff + 1) + low;
    }
    
    public static final void startAutoCalibration(EnchantingCalibratedSensor ld) {
    	int raw_value = ld.getUncalibratedValue();
    	ld.setHigh(raw_value);
    	ld.setLow(raw_value);
    }
    public static final void autoCalibrate(EnchantingCalibratedSensor ld) {
    	int raw_value = ld.getUncalibratedValue();
    	if (ld.getHigh() < raw_value) { ld.setHigh(raw_value); }
    	if (ld.getLow () > raw_value) { ld.setLow (raw_value); }
    }
    
    public static final void setMotorSpeedAndDir(NXTRegulatedMotor m, int speed) {
    	if (speed > 0) {
    		m.forward();
    	} else if (speed < 0) {
    		m.backward(); 
    	}
    	m.setSpeed(speed);
    }
    
    public static final double log10(double x) {
		// return Math.log10(x);	// not implemented
    	// return Math.log(x) / Math.log(10);
    	return Math.log(x) / 2.302585092994046;  // (no sense recalculating log(10))    	
    }
    
    public static final float timer_count()
    {
    	return stopwatch.elapsed() / 1000.0f;
    }
    
    public static final void reset_timer()
    {
    	stopwatch.reset();
    }
	    
    // Returns a number representing the ascii (or Unicode) value of the first
    // element of the string
    public static int AsciiValueOf(String s) {
    	return (int)s.charAt(0);
    }
    
    // Returns a character representation from the inputted number
    public static char AsciiCharacterFrom(int i) {
    	return (char)i;
    }
    
    public static boolean SeesColour(EnchantingColorOrLightSensor sensor, EObject colour_list) {
    	HSV seen_colour = sensor.readColor().asHSV();
    	HSV target_colour = colour_list.asHSV();    	
    	return seen_colour.colorID() == target_colour.colorID();
    }

	public static float GetColorComponent(EObject colour, EObject component) {
		// returns the hue, saturation, or value components of a colour
		// when component is 1, 2, or 3 respectively
		if (component.asNumber() <= 3) {
			return NumberFrom(colour.getItemFromList(component));
		}
		// or the colour ID if component is 4
		return NumberFrom(colour.asHSV().colorID());
	}


	// NAVIGATIONAL COMMANDS for Enchanting v0.2
	
	// Makes the robot start driving forwards or backwards
	public static boolean drive(Navigator navigator, DifferentialPilot pilot, int direction) {
		if (navigator.isMoving()) return false;
		if (direction > 0) {
			pilot.forward();
		} else {
			pilot.backward();
		}
		return true;
	}

	// Makes robot travel (if navigator isn't doing anything)
	// Returns true if movement occurred.
	public static boolean drive(Navigator navigator, DifferentialPilot pilot, int direction, float distance, float units) {
		if (navigator.isMoving()) return false;
		pilot.travel(distance * direction * units);
		return true;
	}

	// Makes robot travel (if navigator isn't doing anything)
	// Returns true if movement occurred.
	public static boolean steer(Navigator navigator, DifferentialPilot pilot, int direction, float turnRate) {
		if (navigator.isMoving()) return false;
		if (direction > 0) {
			pilot.steer(turnRate);
		} else {
			pilot.steerBackward(turnRate);
		}
		return true;
	}

	// Makes robot rotate (if navigator isn't doing anything)	
	// Returns true if rotation occurred.
	public static boolean rotateRobot(Navigator navigator, DifferentialPilot pilot, int direction) {
		if (navigator.isMoving()) return false;
		if (direction > 0) {
			pilot.rotateLeft();
		} else {
			pilot.rotateRight();			
		}
	   return true;
	}

	// Makes robot rotate (if navigator isn't doing anything)	
	// Returns true if rotation occurred.
	// Positive angle rotates left (anti-clockwise), negative right
	public static boolean rotateRobot(Navigator navigator, DifferentialPilot pilot, int direction, float degrees) {
		if (navigator.isMoving()) return false;
		if (pilot instanceof RotateMoveController)
		{
		   ((RotateMoveController) pilot).rotate(direction * degrees);
		   return true;
		}
		return false;
	}

	public static boolean arc(Navigator navigator, DifferentialPilot pilot, int direction, int turn, float radius, float units) {
		if (navigator.isMoving()) return false;
		float r = radius * units * turn;
		if (direction > 0) {
			pilot.arcForward(r);
		} else {
			pilot.arcBackward(r);
		}
		return true;		
	}

	public static boolean arcThrough(Navigator navigator, DifferentialPilot pilot, int direction, int turn, float radius, float units, float angle) {
		if (navigator.isMoving()) return false;
		double _radius = radius * units * turn;  // positive radius turns left
		double _angle = angle * direction * turn; // go forward if radius and angle have same sign
		pilot.arc(_radius, _angle);
		return true;
	}

	// Tells the robot to go to a specific location
	public static void goToXY(Navigator navigator, DifferentialPilot pilot, float x, float y) {
		navigator.goTo(x, y);
		navigator.waitForStop();
	}
	
	// Tells the robot to go to a specific location with a specific heading
	public static void goToXYDirection(Navigator navigator, DifferentialPilot pilot, float x, float y, float dir) {
		navigator.goTo(x, y, enchantingDirectionToLejosHeading(dir));
		navigator.waitForStop();
	}
	
	
	// Used to tell the robot where it actually is
	public static void resetRobotXPosition(Navigator navigator, DifferentialPilot pilot, float x) {
		PoseProvider pp = navigator.getPoseProvider();
		Pose pose = pp.getPose();
		pose.setLocation(x, pose.getY());
		pp.setPose(pose);
	}
	public static void resetRobotYPosition(Navigator navigator, DifferentialPilot pilot, float y) {
		PoseProvider pp = navigator.getPoseProvider();
		Pose pose = pp.getPose();
		pose.setLocation(pose.getX(), y);
		pp.setPose(pose);
	}
	public static void resetRobotDirection(Navigator navigator, DifferentialPilot pilot, float dir) {
		PoseProvider pp = navigator.getPoseProvider();
		Pose pose = pp.getPose();
		pose.setHeading(enchantingDirectionToLejosHeading(dir));
		pp.setPose(pose);
	}

	public static void resetRobotPose(Navigator navigator, DifferentialPilot pilot, float x, float y, float dir) {
		Pose pose = new Pose(x, y, enchantingDirectionToLejosHeading(dir));
		navigator.getPoseProvider().setPose(pose);
	}

	// MISC FUNCTIONS
	
	public static final float map(float value, float low1, float high1, float low2, float high2) {
		float span1 = high1 - low1;
		float span2 = high2 - low2;
		if (span1 == 0 || span2 == 0) return low2;
		return ((value - low1) / span1) * span2 + low2;
	}

    // CONVERSION FUNCTIONS
    // --------------------
    
    // String
    public static final String StringFrom(Object o) {
        return o.toString();
    }
    public static final String StringFrom(EObject e) {
    	return e.asText();
    }
    
    // Number
    public static final float NumberFrom(Object o) {
        return NumberFrom(o.toString());
    }
    public static final float NumberFrom(EObject e) {
    	return e.asNumber();
    }
    public static final float NumberFrom(String s) {
        try {
            return Float.parseFloat(s);
        } catch (NumberFormatException e) {
            return 0;
        }
    }
    public static final float NumberFrom(int i) {
        return (float)i;
    }
    public static final float NumberFrom(float f) {
        return f;
    }
    public static final float NumberFrom(double d) {
        return (float)d;
    }
    public static final float NumberFrom(boolean b) {
        if (b) {
            return 1.0f;
        }
        return 0.0f;
    }
    public static final int ToNearestInt(float f) {
    	return (int)(f + 0.5f);	// round to nearest integer
    }
    
    public static final int IntegerFrom(Object o) {
        return ToNearestInt(NumberFrom(o));
    }
    public static final int IntegerFrom(EObject e) {
    	return ToNearestInt(e.asNumber());
    }
    
    public static final int IntegerFrom(String s) {
    	return ToNearestInt(NumberFrom(s));
    }
    public static final int IntegerFrom(int i) {
    	return i;
    }
    public static final int IntegerFrom(float f) {
    	return ToNearestInt(NumberFrom(f));
    }
    public static final int IntegerFrom(double d) {
    	return ToNearestInt(NumberFrom(d));
    }    	
    public static final int IntegerFrom(boolean b) {
    	return b ? 1 : 0;
    }

	// Boolean
    public static final boolean BooleanFrom(Object o) {
    	//TODO check what type it is and cast accordingly
    	return BooleanFrom(o.toString());
    }
    public static final boolean BooleanFrom(EObject e) {
    	return e.asBoolean();
    }
    
    public static final boolean BooleanFrom(boolean b) {
        return b;
    }
    public static final boolean BooleanFrom(float f) {
        return f != 0.0f;
    }
    public static final boolean BooleanFrom(String s) {
        // A string equates to false if it is empty, "false", or "0"
        if (s == "") return false;
        if (s.equalsIgnoreCase("false")) return false;
        try {
            if (Float.parseFloat(s) == 0.0f) return false;
        } catch (NumberFormatException e) {}
        
        return true;
    }
    
    // EObject
    public static final EObject EObjectFrom(Object o) {
    	if (o instanceof EObject)
    		return (EObject) o;
    	else
    		return new EObject(o);
    }
    public static final EObject EObjectFrom(EObject e) {
    	return e;
    }
    
    // Command
/**	public static final Command CommandFrom(Object o) {
    	if (o instanceof Command)
    		return (Command)o;
    	else
    		return new Command();
    }
    public static final Command CommandFrom(EObject e) {
    	return e.asCommand();
    }
    
    // Reporter
    public static final Reporter ReporterFrom(Object o) {
        if (o instanceof Reporter)
        	return (Reporter)o;
		else if (o instanceof EObject)
			return ReporterFrom((EObject)o);
		else
			return new Reporter();
    }
    public static final Reporter ReporterFrom(EObject e) {
    	return e.asReporter();
    }
    
    // Predicate
    public static final Predicate PredicateFrom(Object o) {
        if (o instanceof Predicate)
        	return (Predicate)o;
        else if (o instanceof EObject)
        	return PredicateFrom((EObject)o);
		else
			return new Predicate();
    }
    public static final Predicate PredicateFrom(EObject e) {
    	return e.asPredicate();
    }*/
    
    public static final CustomBlock CustomBlockFrom(Object o) {
    	if (o instanceof CustomBlock)
    		return (CustomBlock)o;
    	else if (o instanceof EObject)
    		return CustomBlockFrom((EObject)o);
    	else
    		return new Command();
    }
    public static final CustomBlock CustomBlockFrom(EObject e) {
    	return e.asCustomBlock();
    }
    
    // In all of these comparison functions, we first assume that
    // we are comparing two numbers.  If we can't get two numbers to compare,
    // we compare strings.  [Comparing bools doesn't really make sense.] 
    /**
    public static final boolean GreaterThan(String a, String b) {
        try {
            return Float.parseFloat(a) > Float.parseFloat(b);
        } catch (NumberFormatException e) { }
        return CompareStrings(a, b) == +1;
    }
    public static final boolean LessThan(String a, String b) {
        try {
            return Float.parseFloat(a) < Float.parseFloat(b);
        } catch (NumberFormatException e) { }
        return CompareStrings(a, b) == -1;
    }
    public static final boolean EqualTo(String a, String b) {
        try {
            return Float.parseFloat(a) == Float.parseFloat(b);
        } catch (NumberFormatException e) { }
        //return a.equalsIgnoreCase(b);        
        return CompareStrings(a, b) == 0;
    }
    public static final int CompareStrings(String a, String b) {
        // Compares two strings, ignoring case.
        // -1 means a < b, 0 means a == b, and +1 means a > b
        // Probably a slow function.
        String a2 = a.toLowerCase();
        String b2 = b.toLowerCase();
        int lenA = a2.length();
        int lenB = b2.length();
        for (int i = 0; i < lenA && i < lenB; ++i) {
            if (a2.charAt(i) != b2.charAt(i))
                return (a2.charAt(i) > b2.charAt(i)) ? +1 : -1;
        }
        if (lenA == lenB) return 0;
        return (lenA > lenB) ? +1 : -1; // longer string is greater
    }
    */
    
    // In Enchanting, we've decided to mimic the sprite's orientation.
    // The heading as used by leJOS is different.
    // (Compare these to the heading functions for the sprites).
    
    //          	 			Enchanting     	leJOS
    // Range:	    			-180 to +180   	0 to 360
    // Pointing
    //   along +x / "right":		 90			  0          -90		270
    //	 along +y / "up":			  0			 90            0		360
    //	 along -x / "left":			-90			180          -270		 90
    //   along -y / "down":			180			270          -180  		180
        
    public static float enchantingDirectionToLejosHeading(float dir) {
    	float heading = (360.0f + 90.0f - dir) % 360.0f;
    	return heading;
    }
    public static float lejosHeadingToEnchantingDirection(float heading) {
    	float dir = 90.0f - (heading % 360.0f);
    	if (dir <= -180.0f) { dir += 360.0f; }
    	return dir;
    }
        
}
